<?php
namespace Ekliptor\Wordpress;

use Ekliptor\CashP\BlockchainApi\AbstractBlockchainApi;
use Ekliptor\CashP\BlockchainApi\Http\WordpressHttpAgent;

require_once BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/autoload.php';
require_once BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/Settings.php';
require_once BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/Sanitizer.php';
// TODO move some classes to Admin class if we really don't need them here
require_once BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/TemplateEngine.php';

require_once BCHSLP_PAYMENTS__PLUGIN_DIR . 'classes/AdminNotice.php';
include BCHSLP_PAYMENTS__PLUGIN_DIR . 'structs/PreInstalledTokens.php';

/**
 * The class containing the core functions of this plugin, handling actiation/deactivaton, loading state/settings,...
 */
class SlpPaymentsCore {
	const DEBUG = true;
	const SESSION_LIFETIME_SEC = 365 * DAY_IN_SECONDS;
	const CLEANUP_TRANSACTIONS_H = 25;
	const BCH_TX_FEE = array(
			'USD' => 0.021 // use a higher value
	);
	const BCH_TX_FEE_TOLERANCE = 0.00075;
	const BCH_DUST_LIMIT_SAT = 546; // https://github.com/Bitcoin-ABC/bitcoin-abc/blob/9bbaacc1e0d86a7a18352068b93ca0ccab45f2ea/src/test/transaction_tests.cpp#L546
	const BADGER_WALLET_JS = "https://developer.bitcoin.com/badger/badgerButton-1.0.1.js";
	const MAX_AGE_TX_ACCEPT = 15*10; // max 15 blocks older
	
	/** @var SlpPaymentsCore */
	private static $instance = null;
	private static $prefix = "slp_";
	private static $cron_events = array (
			'slp_cleanup_transactions'
		);
	private static $cron_events_hourly = array(
			'slp_update_currency_rates',
			//'slp_check_transaction_confirmations' // moved to SlpPayments
		);
	/** @var \WC_Log_Handler_File */
    protected static  $logger = null;
	
	/** @var Settings */
	protected $settings;
	/** @var array URL parts of this WP site as the result of: parse_url(site_url('/')) */
	protected $siteUrlParts = null;
	/** @var Sanitizer */
	protected $sanitizer = null;
	
	/** @var bool */
	protected $includedButtonScript = false;
	/** @var string */
	protected $qrcodeStatic;
	/** @var string */
	protected $closeImage;
	/** @var string */
	protected $loadingImage;
	
	private function __construct() {
		$this->qrcodeStatic = plugins_url( 'img/qrcode.svg', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		$this->closeImage = plugins_url( 'img/close50.png', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		$this->loadingImage = plugins_url( 'img/loading.gif', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		$this->init();
	}
	
	public static function getInstance() {
		if (self::$instance === null)
			self::$instance = new self ();
		return self::$instance;
	}
	
	public static function getTableName($tableName) {
		global $wpdb;
		return $wpdb->prefix . self::$prefix . $tableName;
	}
	
	public function init() {
		$this->siteUrlParts = parse_url(site_url('/'));
		if (class_exists('SlpPaymentsAdmin', false) === true) // must be delayed until init call so that source is ready
			$this->settings = new Settings($this, SlpPaymentsAdmin::getInstance($this));
		else
			$this->settings = new Settings($this);
		$availableTokens = $this->settings->get('availableTokens');
		if (empty($availableTokens) || !isset($availableTokens['USDH']) || !isset($availableTokens['USDH']['name'])) // first install or config has been deleted
			$this->settings->set('availableTokens', PreInstalledTokens::getSettingsArray());
		$this->sanitizer = new Sanitizer();
		
		// init hooks
		// note that functions must be public because they are called from event stack
		add_action( 'wp_enqueue_scripts', array ($this, 'addPluginScripts' ) );
		add_action( 'wp_footer', array($this, 'addFooterCode') );
		
		// Crons
		//add_action ( 'slp_cleanup_transactions', array ($this, 'cleanupTransactions' ) ); // must be scheduled on an earlier hook
		//add_action ( 'slp_update_currency_rates', array ($this, 'updateCurrencyRates' ) );
		
		if ($this->settings->get('rate_usd_bch') === 0.0)
			add_action ( 'shutdown', array ($this, 'updateCurrencyRates' ) );
		else if ($this->settings->get('last_rate_update') + 90*MINUTE_IN_SECONDS < time())
			add_action ( 'shutdown', array ($this, 'triggerCrons' ) );
	}
	
	public static function plugin_activation() {
		if (version_compare ( $GLOBALS ['wp_version'], BCHSLP_PAYMENTS__MINIMUM_WP_VERSION, '<' )) {
			load_plugin_textdomain ( 'ekliptor' );
			$message = '<strong>' . sprintf ( esc_html__ ( 'SLP Payments plugin %s requires WordPress %s or higher.', 'ekliptor' ), BCHSLP_PAYMENTS_VERSION, BCHSLP_PAYMENTS__MINIMUM_WP_VERSION ) . '</strong> ' . sprintf ( __ ( 'Please <a href="%1$s">upgrade WordPress</a> to a current version.', 'ekliptor' ), 'https://codex.wordpress.org/Upgrading_WordPress' );
			static::bailOnActivation ( $message, false );
		}
		
		$transactionsTableSuccess = static::createTransactionsTable();
		if ($transactionsTableSuccess === false) {
			load_plugin_textdomain ( 'ekliptor' );
			$message = '<strong>' . esc_html__ ( 'Error creating required MySQL tables.', 'ekliptor' ) . '</strong> ' . sprintf ( __ ( 'Please ensure that you have sufficient database privileges. If you still encounter this problem afterwards, please <a href="%1$s">file a bug report</a>.', 'ekliptor' ), 'https://simpleledger.io/' );
			static::bailOnActivation ( $message, false );
		}
		$tables = get_option('slp_tables', array());
		if (in_array(static::getTableName('transactions'), $tables) === false)
			$tables[] = static::getTableName('transactions');
		update_option('slp_tables', $tables);
		
		// ensure directories exist
		$dataDirs = array(
				BCHSLP_PAYMENTS__PLUGIN_DIR . 'data',
				BCHSLP_PAYMENTS__PLUGIN_DIR . 'data/temp',
				BCHSLP_PAYMENTS__PLUGIN_DIR . 'data/temp/qr'
		);
		foreach ($dataDirs as $dir) {
			if (file_exists($dir) === true)
				continue;
			if (mkdir($dir) === false) { // TODO even though we don't create php files, using WP filesystem API would still be better
				load_plugin_textdomain ( 'ekliptor' );
				$message = '<strong>' . esc_html__ ( 'Error creating data folder.', 'ekliptor' ) . '</strong> ' . sprintf ( __ ( 'Please ensure that your WordPress installation has write permissions on the /plugins folder (0755) or create this folder manually with permissions 0755: %s', 'ekliptor' ), $dir );
				static::bailOnActivation ( $message, false );
			}
		}
		
		foreach ( self::$cron_events as $cron_event ) {
			$timestamp = wp_next_scheduled ( $cron_event );
			if (!$timestamp)
				wp_schedule_event(time(), 'daily', $cron_event);
		}
		foreach ( self::$cron_events_hourly as $cron_event ) {
			$timestamp = wp_next_scheduled ( $cron_event );
			if (!$timestamp)
				wp_schedule_event(time(), 'hourly', $cron_event);
		}
		
		//$lastVersion = get_option ( 'slp_payments_installed' );
		//update_option ( 'slp_payments_installed', BCHSLP_PAYMENTS_VERSION ); // done from WC class
	}
	
	public static function plugin_deactivation() {
		//global $wpdb;
		// Remove any scheduled cron jobs.
		$events = array_merge(self::$cron_events, self::$cron_events_hourly);
		foreach ( $events as $cron_event ) {
			$timestamp = wp_next_scheduled ( $cron_event );
			if ($timestamp) {
				wp_unschedule_event ( $timestamp, $cron_event );
			}
		}
		
		// tables are only being dropped on uninstall. also plugin settings
		
		// delete_option('slp_payments_installed');
	}
	
	public static function setupCronHooks() {
		add_action ( 'slp_cleanup_transactions', array (static::getInstance(), 'cleanupTransactions' ) );
		add_action ( 'slp_update_currency_rates', array (static::getInstance(), 'updateCurrencyRates' ) );
		add_action ( 'slp_unused_address_search', array (static::getInstance(), 'searchUnusedAddress' ) );
		
		/*
		define('DOING_CRON', true);
		$hookname = 'slp_update_currency_rates';
		$args = array();
		delete_transient( 'doing_cron' );
		wp_schedule_single_event( time() - 1, $hookname, $args );
		if (spawn_cron() != true)
			echo "failed cron";
		*/
		//do_action('slp_update_currency_rates');
	}
	
	public function toAmountBCH(float $amount, string $currency = ''): float {
		if ($currency === '')
			$currency = $this->settings->get('button_currency');
		if ($currency !== 'USD')
			return 0.0; // TODO crawl other rates and add support for more currencies
		$rate = $this->settings->get('rate_usd_bch');
		if ($rate === 0.0)
			return 0.0;
		return floor($amount / $rate * 100000000) / 100000000; // many wallets support only 8 decimals
	}
	
	public function addFooterCode() {
		$cfg = array(
			'cookieLifeDays' => ceil(static::SESSION_LIFETIME_SEC / DAY_IN_SECONDS),
			'cookiePath' => $this->siteUrlParts['path'],
			'siteUrl' => $this->getSiteUrl(),
			'display_currency' => strtolower($this->settings->get('button_currency')),
			'display_digits_bch' => $this->settings->get('display_digits_bch'),
			'rate' => array(
					'usd' => $this->settings->get('rate_usd_bch')
			),
			'slpSocketAddress' => $this->settings->get('slpSocketAddress'),
			/*
			'tokenRate' => array(
					'usd' => $this->settings->get('token_rate_usd')
			),
			*/
			// TODO add multi fialt currently support. currently we assume same currency as in WC
			'availableTokens' => $this->settings->get('availableTokens'),
			// TODO move localized strings into a separate .js file generated from PHP if we have more strings
			'badgerLocked' => __('Your BadgerWallet is locked. Please open it in your browser toolbar and enter your password before sending money.', 'ekliptor'),
		);
		$cfg = apply_filters('slp_payments_js_config', $cfg);
		echo '<script type="text/javascript">var slpPaymentsCfg = ' . json_encode($cfg) . ';</script>';
	}
	
	public function addPluginScripts() {
		wp_enqueue_style( 'slp-payments', plugins_url( 'tpl/css/slp-payments.css', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' ), array(), BCHSLP_PAYMENTS_VERSION );
		wp_enqueue_script( 'slp-payments-bundle', plugins_url( 'tpl/js/bundle.js', BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' ), array(), BCHSLP_PAYMENTS_VERSION, true );
	}
	
	public function getSettings(): Settings {
		return $this->settings;
	}
	
	public function useTokenPayments(): bool {
		//return $this->settings->get('enabledToken') === true && !empty($this->settings->get('tokenID'));
		return count($this->getEnabledTokens()) > 0;
	}
	
	public function cleanupTransactions() {
		global $wpdb;
		$table = static::getTableName('transactions');
		$maxAge = date('Y-m-d H:i:s', time() - static::CLEANUP_TRANSACTIONS_H * HOUR_IN_SECONDS);
		$wpdb->query("DELETE FROM $table WHERE created < '$maxAge'");
		
		// cleanup the dir with QR codes
		$cacheDir = BCHSLP_PAYMENTS__PLUGIN_DIR . 'data/temp/qr/';
		$files = scandir($cacheDir);
		if ($files === false) {
			static::notifyErrorExt('Error scanning qr code dir to cleanup', "cache dir: $cacheDir");
			return;
		}
		// cleanup by age, oldest creation/changed time first
		$deletionTime = time() - static::CLEANUP_TRANSACTIONS_H*HOUR_IN_SECONDS;
		foreach ($files as $file) {
			if (empty($file) || $file[0] === '.')
				continue;
			$filePath = $cacheDir . '/' . $file;
			$lastChanged = filectime($filePath);
			if ($lastChanged < $deletionTime)
				@unlink($filePath);
		}
	}
	
	public function updateCurrencyRates() {
		$res = wp_remote_get('https://index-api.bitcoin.com/api/v0/cash/price/usd');
		if (is_array($res) === false) {
			static::notifyErrorExt('Error updating currency rates', $res);
			return;
		}
		$json = json_decode($res['body']);
		if ($json === null || empty($json->price)) {
			static::notifyErrorExt('Invalid response when updating currency rates', $res);
			return;
		}
		if ($json->price > 0.0) {
			$this->settings->set('rate_usd_bch', $json->price / 100.0); // receiced in USD cents
			$this->settings->set('last_rate_update', time());
		}
	}
	
	public function scheduleUnsuedAddressSearch($offsetSec = 0): int {
		$scheduleTime = time() + $offsetSec;
		wp_schedule_single_event($scheduleTime, 'slp_unused_address_search');
		return $scheduleTime;
	}
	
	public function searchUnusedAddress() {
		$scheduleTime = $this->scheduleUnsuedAddressSearch(5*MINUTE_IN_SECONDS); // schedule it again in case it gets aborted
		$xPub = $this->settings->get('xPub');
		$hdPathFormat = $this->settings->get('hdPathFormat');
		$nextCount = $this->settings->get('addressCount') + 1; // in live mode we also increase +1 when creating the 1st address
		$skipCount = $this->settings->get('skipUsedAddressCount');
		if ($skipCount < 1)
			$skipCount = 1;
		$blockchainApi = $this->createBlockchainApiInstance();
		$hasExistingTransactions = false;
		while ($hasExistingTransactions === false)
		{
			// make sure we don't timeout
			// WP cron already calls ignore_user_abort
			set_time_limit(60);
			$address = $blockchainApi->createNewAddress($xPub, $nextCount, $hdPathFormat);
			if (!$address || !$address->cashAddress) { // try again later
				static::notifyError("Failed to create address when finding unused address", "");
				return;
			}
			$balance = $blockchainApi->getAddressBalance($address->cashAddress);
			if ($balance > 0.0)
				$nextCount += $skipCount;
			else {
				$this->settings->set('addressCount', $nextCount);
				break; // we have succeeded
			}
		}
		//static::notifyError("found empty address $nextCount", $address);
		wp_unschedule_event($scheduleTime, 'slp_unused_address_search'); // done successfully. remove it from schedule
	}
	
	public function createBlockchainApiInstance(string $implementation = '', string $blockchainApiUrl = ''): AbstractBlockchainApi {
		if (empty($implementation) || empty($blockchainApiUrl)) {
			$implementation = $this->settings->get('blockchain_api');
			$blockchainApiUrl = $this->settings->get('blockchain_rest_url');
		}
		$blockchainApi = AbstractBlockchainApi::getInstance($implementation, $blockchainApiUrl);
		AbstractBlockchainApi::setLogger(function (string $subject, $error, $data = null) {
			SlpPaymentsCore::notifyErrorExt("BlockChain API: " . $subject, $error, $data);
		});
		$httpAgent = new WordpressHttpAgent(function (string $subject, $error, $data = null) {
			SlpPaymentsCore::notifyErrorExt("HTTP: " . $subject, $error, $data);
		});
		$blockchainApi->setHttpAgent($httpAgent);
		return $blockchainApi;
	}
	
	public function notifyError($subject, $error, $data = null) {
		static::notifyErrorExt($subject, $error, $data); // member function included for backwards compatibility
	}
	
	public static function notifyErrorExt($subject, $error, $data = null) {
		global $wpdb;
		if (defined('WC_LOG_DIR') === true) {
			if (static::$logger === null)
				static::$logger = new \WC_Log_Handler_File();
			$logMsg = $subject . "\r\n" . print_r($error, true);
			if ($data !== null)
				$logMsg .= "\r\n" . print_r($data, true);
			static::$logger->handle(time(), 'error', $logMsg, array('source' => 'slp-payments'));
		}
		if (static::DEBUG === false)
			return;
		$table = static::getTableName("messages_system");
		if ($wpdb->get_var("SHOW TABLES LIKE '$table'") !== $table) {
			pre_print_r($subject);
			pre_print_r($error);
			if ($data !== null)
				pre_print_r($data);
			return; // table doesn't exist
		}
		if (!is_string($error))
			$error = print_r($error, true);
		$rowCount = $wpdb->insert($table, array(
				'sender' => 'SystemError',
				'subject' => $subject,
				'text' => $error,
				'data' => $data !== null ? serialize($data) : null,
				'site' => strtolower(get_bloginfo('name'))
		));
	}
	
	public static function getRandomString($len) {
		$chars = '1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
		$max = strlen($chars)-1;
		mt_srand();
		$random = '';
		for ($i = 0; $i < $len; $i++)
			$random .= $chars[mt_rand(0, $max)];
		return $random;
	}
	
	public static function toSatoshis($bch) {
		return floor($bch * 100000000);
	}
	
	public static function isValidUrl(string $url): bool {
		if (strlen($url) < 6 || preg_match("/^https?:\/\//", $url) !== 1 || strpos($url, '.') === false)
			return false;
		return true;
	}
	
	public function getSiteUrl(array $query = array()) {
		$url = $this->siteUrlParts['scheme'] . '://' . $this->siteUrlParts['host'];
		if (isset($this->siteUrlParts['port']))
			$url .= $this->siteUrlParts['port'];
		$url .= $this->siteUrlParts['path'];
		$first = true;
		foreach ($query as $key => $value) {
			$url .= $first === true ? '?' : '&';
			$url .= $key . '=' . urlencode($value);
			$first = false;
		}
		return $url;
	}
	
	public function getCurrentUrl() {
		global $wp;
		return home_url( add_query_arg( array(), $wp->request ) );
	}
	
	public function getIncludedButtonScript(): bool {
		return $this->includedButtonScript;
	}
	
	public function setIncludedButtonScript(bool $included) {
		$this->includedButtonScript = $included;
	}
	
	public function getEnabledTokens(): array {
		$enabledTokens = array();
		$availableTokens = $this->settings->get('availableTokens');
		foreach ($availableTokens as $symbol => $token) {
			if (!empty($token['id']) && $token['rate'] > 0.0 && $token['enabled'] === true)
				$enabledTokens[$symbol] = $token;
		}
		return $enabledTokens;
	}
	
	/**
	 * Returns the SLP token with the given symbol as associative array or null if this token doesn't exist in our local database.
	 * @param string $symbol
	 * @return array|NULL
	 */
	public function getTokenBySymbol(string $symbol)  {
		$availableTokens = $this->settings->get('availableTokens');
		return isset($availableTokens[$symbol]) ? $availableTokens[$symbol] : null;
	}
	
	/**
	 * On some installations WP cron might not run properly. As a workaround we use this function which
	 * runs in the main script. You should attach it the shutdown action to ensure
	 * all action hooks have been attached previously and the user-output is generated first.
	 */
	public function triggerCrons() {
		static $doingCrons = false;
		if ($doingCrons === true)
			return;
		$doingCrons = true;
		do_action('slp_check_transaction_confirmations');
		do_action('slp_cleanup_transactions');
		do_action('slp_update_currency_rates');
	}
	
	/* // same addresses for all users/posts generated by xpub
	public function getReceiverAddress($attrs = array()) {
		if ($this->settings->get('author_bch_addresses') === false) 
			return $this->settings->get('bch_address');
		$address = isset($attrs['address']) && $this->sanitizer->isValidBitcoinCashAddress($attrs['address']) ? trim($attrs['address']) : $this->settings->get('bch_address');
		$authorID = (int)get_the_author_meta('ID');
		if ($authorID === 0)
			return $address;
		$authorAddress = get_user_meta($authorID, 'author_bch_address', true);
		return $authorAddress ? $authorAddress : $address;
	}
	*/
	
	/*
	public function createNewReceiverAddress($attrs = array()): string {
		$nextCount = $this->settings->get('addressCount') + 1;
		$this->settings->set('addressCount', $nextCount);
		$bchAddress = $this->blockchainApi->createNewAddress($this->settings->get('xPub'), $nextCount, $this->pluginSettings->get('hdPathFormat'));
		$this->settings->set('lastAddress', $bchAddress->cashAddress);
		return '';
	}
	*/
	
	/**
	 * Generate a QR code for the payment
	 * @param string $txid The internal MySQL transaction ID (not the on-chain TXID).
	 * @param string $address The (1-time) BCH address created for this payment.
	 * @param float $amount The amount to be received in display currency (USD).
	 * @param float (optional) The amount in BCH. Required if the user pays part of the amount in SLP tokens. 
	 * @param float $amountToken (optional) The amount of the optional token to be received.
	 * @param string $tokenID (optional) The hex ID of the SLP token. Required if $amountToken > 0.
	 * @param int $tokenDigits (optional). The number of decimal places to use in the payment URI of this token.
	 * @return string the public image URL of the QR code
	 */
	public function generateQrCodeForAddress(string $txid, string $address, float $amount, float $amountBCH = 0.0, float $amountToken = 0.0, string $tokenID = "", int $tokenDigits = 8): string {
		$qrHash = hash('sha512', $txid);
		$fileName = sprintf('data/temp/qr/%s.png', $qrHash);
		$fileLocal = BCHSLP_PAYMENTS__PLUGIN_DIR . $fileName;
		$fileUrl =  plugins_url( $fileName, BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		if (file_exists($fileLocal) === true)
			return $fileUrl; // use it from cache
		
		if ($amountBCH == 0.0)
			$amountBCH = $this->toAmountBCH($amount, $this->settings->get('button_currency'));
		$codeContents = $this->createPaymentURI($address, $amountBCH, $amountToken, $tokenID, $tokenDigits);
		\QR_Code\QR_Code::png($codeContents, $fileLocal);
		return $fileUrl;
	}
	
	/**
	 * Generate a QR code for the payment
	 * @param string $txid The internal MySQL transaction ID (not the on-chain TXID).
	 * @param float (optional) The amount in BCH. Required if the user pays part of the amount in SLP tokens. 
	 * @return string the public image URL of the QR code
	 */
	public function generateQrCode(string $txid, float $amountBCH = 0.0): string {
		global $wpdb;
		$qrHash = hash('sha512', $txid);
		$fileName = sprintf('data/temp/qr/%s.png', $qrHash);
		$fileLocal = BCHSLP_PAYMENTS__PLUGIN_DIR . $fileName;
		$fileUrl =  plugins_url( $fileName, BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
		if (file_exists($fileLocal) === true)
			return $fileUrl; // use it from cache
		
		$table = SlpPaymentsCore::getTableName('transactions');
		$query = $wpdb->prepare("SELECT address, amount, amountToken, tokenSymbol, tokenID FROM $table WHERE txid = '%s'", array($txid));
		$row = $wpdb->get_row($query);
		if (empty($row))
			return '';
		$slpToken = $this->getTokenBySymbol($row->tokenSymbol);
		$decimals = 8;
		if ($slpToken && isset($slpToken['decimals']))
			$decimals = $slpToken['decimals'];
		return $this->generateQrCodeForAddress($txid, $row->address, $row->amount, $amountBCH, $row->amountToken, $row->tokenID, $decimals);
	}
	
	/**
	 * Return a payment URI (starting with "bitcoincash:" or "simpleledger:" if $amountToken > 0) for the given $address.
	 * @param string $address
	 * @param float $amountBCH
	 * @param float $amountToken
	 * @param string $tokenID
	 * @param int $tokenDigits
	 * @return string
	 */
	public function createPaymentURI(string $address, float $amountBCH, float $amountToken = 0.0, string $tokenID = "", int $tokenDigits = 8): string {
		if ($tokenDigits < 0)
			$tokenDigits = 0;
		$address = preg_replace("/.+:/i", "", $address);
		// we use the bitcoincash URI if tokens are disabled because bitcoin.com and other wallets only support this as of June 2019
		//$uri = "simpleledger:$address?amount=$amountBCH";
		$uri = sprintf("bitcoincash:%s?amount=%s", $address, number_format($amountBCH, $tokenDigits, ".", ""));
		if ($amountToken > 0.0 && /*$this->useTokenPayments() === true*/!empty($tokenID)) {
			//$tokenID = $this->settings->get('tokenID');
			if ($amountBCH > 0.0)
				$uri = sprintf("simpleledger:%s?amount=%s&", $address, number_format($amountBCH, $tokenDigits, ".", ""));
			else
				$uri = sprintf("simpleledger:%s?", $address);
			$uri .= sprintf("amount1=%s-%s", number_format($amountToken, $tokenDigits, ".", ""), $tokenID);
		}
		return $uri;
	}
	
	/**
	 * Create a new unique internal transaction ID and store it in our database to keep track of this pament.
	 * @param string $recAddress The receiving BCH address for this payment.
	 * @param float $amount The amount in display currency (USD).
	 * @param float $amountToken The amount in the optional token.
	 * @param int $postID The WP post ID or WC order ID this transaction belongs to.
	 * @param string $tokenID
	 * @param string $tokenSymbol
	 * @return boolean|string
	 */
	public function createTransactionId(string $recAddress, float $amount, float $amountToken = 0.0, int $postID = 0, string $tokenID = "", string $tokenSymbol = "") {
		global $wpdb;
		$table = static::getTableName('transactions');
		$txid = static::getRandomString(40);
		$data = array(
				'txid' => $txid,
				'address' => $recAddress,
				'amount' => $amount,
				'amountToken' => $amountToken,
				'tokenID' => $tokenID,
				'tokenSymbol' => $tokenSymbol,
				'post_id' => $postID
		);
		if ($wpdb->insert($table, $data) !== 1)
			return false; // otherwise we can't check if we received the payment
		return $txid;
	}
	
	public function getRemainingBchAmount(float $tokenAmount, float $totalAmountFiat, string $tokenSymbol = ""): float {
		$currency = strtolower($this->settings->get('button_currency'));
		$bchRates = $this->getBchRates();
		$remainingAmountBase = max(0.0, $totalAmountFiat);
		if (empty($tokenSymbol)) {
			$tokenAmount = 0.0; // should already be
		}
		else {
			$tokenRates = $this->getTokenRate($tokenSymbol);
			$remainingAmountBase = max(0.0, $totalAmountFiat - $tokenAmount*$tokenRates[$currency]);
		}
		$remainingAmountBch = $remainingAmountBase/$bchRates[$currency];
		if (static::toSatoshis($remainingAmountBch) <= static::BCH_DUST_LIMIT_SAT)
			return 0.0;
		return $remainingAmountBch;
	}
	
	public function getTokenAmount(float $fiatAmount, string $tokenSymbol): float {
		$currency = strtolower($this->settings->get('button_currency'));
		$tokenRates = $this->getTokenRate($tokenSymbol);
		$tokenAmount = $fiatAmount / $tokenRates[$currency];
		return $tokenAmount;
	}
	
	public function isSlpAddress(string $address): bool {
		return strtolower(substr($address, 0, 13)) === 'simpleledger:';
	}
	
	/**
	 * Return conversion rates for BitcoinCash (BCH) to fiat currencies.
	 * @return array associative array with keys such as 'usd'
	 */
	public function getBchRates(): array {
		return array(
				'usd' => $this->settings->get('rate_usd_bch')
		);
	}
	
	/**
	 * Return conversion rates for the token to fiat currencies.
	 * @return array associative array:
	 * array('TOKEN_SYMBOL' => array('usd' => ...) )
	 */
	public function getTokenRates(): array {
		$rates = array();
		$availableTokens = $this->settings->get('availableTokens');
		foreach ($availableTokens as $symbol => $token) {
			$rates[$symbol] = array('usd' => $token['rate']);
		}
		return $rates;
	}
	
	/**
	 * Return the fiat token rates of a specific SLP token.
	 * @param string $tokenSymbol
	 * @return array array('usd' => ...)
	 */
	public function getTokenRate(string $tokenSymbol): array {
		$allRates = $this->getTokenRates();
		return isset($allRates[$tokenSymbol]) ? $allRates[$tokenSymbol] : array('usd' => 0.0);
	}
	
	/**
	 * Generate a callback URL with the given arguments.
	 * @param array $args key-value pairs of args as associative array
	 * @return string
	 */
	protected function getWebhookUrl(array $args): string {
		$url = urlencode(site_url("wp-json/v1/slp-payments"));
		$data = '';
		foreach ($args as $key => $value) {
			if (!empty($data))
				$data .= '&';
			$data .= $key . '=' . $value;
		}
		if ($data !== '')
			$url .= '&data=' . urlencode($data); // key and value encoded and added as query param
		return $url;
	}
	
	protected static function createTransactionsTable(): bool {
		global $wpdb;
		$table = static::getTableName('transactions');
		if ($wpdb->get_var("SHOW TABLES LIKE '$table'") === $table)
			return true; // table already exists
		$success = $wpdb->query("CREATE TABLE `$table` (
  				`txid` char(40) NOT NULL,
				`address` varchar(256) NOT NULL,
				`amount` float NOT NULL,
				`amountToken` float NOT NULL,
				`tokenSymbol` varchar(8) NOT NULL,
				`tokenID` char(64) NOT NULL,
				`post_id` bigint(20) NOT NULL,
  				`created` timestamp NOT NULL DEFAULT current_timestamp()
				) ENGINE=InnoDB DEFAULT CHARSET=utf8;") !== false;
		if ($success === false)
			return false;
		$success = $wpdb->query("ALTER TABLE `$table`
  				ADD PRIMARY KEY (`txid`);") !== false;
		return $success === true;
	}
	
	protected static function bailOnActivation($message, $escapeHtml = true, $deactivate = true) {
		include BCHSLP_PAYMENTS__PLUGIN_DIR . 'tpl/message.php';
		if ($deactivate) {
			$plugins = get_option ( 'active_plugins' );
			$cashtippr = plugin_basename ( BCHSLP_PAYMENTS__PLUGIN_DIR . 'slp-payments.php' );
			$update = false;
			foreach ( $plugins as $i => $plugin ) {
				if ($plugin === $cashtippr) {
					$plugins [$i] = false;
					$update = true;
				}
			}
			
			if ($update) {
				update_option ( 'active_plugins', array_filter ( $plugins ) );
			}
		}
		exit ();
	}
}
?>