import {Tooltips} from "./admin/Tooltips";
import {WebHelpers, WebHelpersConfig} from "./WebHelpers";
import {BadgerWallet} from "./BadgerWallet";
import {Woocommerce} from "./Woocommerce";
import {AdminControls} from "./admin/AdminControls";
import {SlpTokenMap} from "./structs/SlpToken";

export interface BitcoinCashConversionRate {
    [fiatCurrency: string]: number;
}
export interface SlpPaymentsConfig extends WebHelpersConfig {
    // same keys as in php
    display_currency: string;
    display_digits_bch: number;
    rate: BitcoinCashConversionRate;
    //tokenRate: BitcoinCashConversionRate; // we now support multiple tokens
    availableTokens: SlpTokenMap;
    totalCartValue: number; // in WooCommerce currency, such as USD
    ajaxConfirm?: boolean;
    keepTransaction?: boolean; // keep the transaction in mysql so that plugin addons can use them after payment
    enabledTokens?: SlpTokenMap;

    // present after order is placed
    orderID?: number;
    nonce?: string;
    tokenID?: string;
    address?: string;
    slpSocketAddress: string;
    checkPaymentIntervalSec: number;
    successPayImage: string; // image url, animated
    successPayImageFinal: string; // image url

    // localizations
    badgerLocked: string;
    paidTxt: string;
}

export interface AbstractPayment {
    domID: string; // the full ID of the wrapper div for the payment control
    txid: string; // the blockchain transaction ID/hash
    address: string; // the receiving address on chain
    amount: number; // the amount in the website's currency (USD,...)
    tokenAmount?: number; // optional: the amount of tokens paid
}

export interface SlpPaymentsApiRes {
    error: boolean;
    errorMsg: string;
    data: any[];
}

export class SlpPayments {
    // TODO separate entryPoints + classes for admin + public code? but tooltips and other admin stuff can be used publicly too (and is quite small)

    public readonly window: Window;
    public readonly $: JQueryStatic;
    public readonly badger: BadgerWallet;
    public readonly woocommerce: Woocommerce;

    protected config: SlpPaymentsConfig;
    protected webHelpers: WebHelpers;
    protected tooltips = new Tooltips(this);
    protected adminControls: AdminControls = null;

    constructor(window: Window, $: JQueryStatic) {
        this.window = window;
        this.$ = $;
        this.config = this.window['slpPaymentsCfg'] || {};
        this.config.consentCookieName = ""; // we currently don't set any additional cookies (other than WP and WC cookies)
        this.config.confirmCookiesMsg = "";
        this.config.confirmCookiesBtn = "";

        this.webHelpers = new WebHelpers(this.window, this.$, this.config);
        this.badger = new BadgerWallet(this, this.webHelpers, true);
        this.woocommerce = new Woocommerce(this, this.webHelpers);
        this.$(this.window.document).ready(($) => {
            this.webHelpers.checkCookieConsent();
            if (this.$("body").hasClass("wp-admin") === true) { // class is added after page load
                this.adminControls = new AdminControls(this);
                this.adminControls.init();
            }
        });
    }

    public static toSatoshis(bch: number) {
        return Math.floor(bch * 100000000);
    }

    public getConfig() {
        return this.config;
    }

    public getTooltips() {
        return this.tooltips;
    }

    public getWebHelpers() {
        return this.webHelpers;
    }

    public getPluginPaymentID(paymentControlsWrapper: JQuery) {
        return paymentControlsWrapper.attr("id").replace(/^slp-btn-wrap-/, "");
    }

    // ################################################################
    // ###################### PRIVATE FUNCTIONS #######################
}
