import {SlpPayments} from "../SlpPayments";
import {SlpToken, SlpTokenMap} from "../structs/SlpToken";
import {AbstractModule} from "../AbstractModule";

export interface SlpPaymentsAdminConfig {
    preInstalledTokens: SlpTokenMap;
    availableTokens: SlpTokenMap;

    // localizations
    tr: {
        newToken: string;
        confirmDeleteToken: string;
        missingFields: string;
    }
}

export class AdminControls extends AbstractModule {
    // add a CSS class in PHP if this selector becomes longer...
    protected static readonly TOKEN_INPUT_SELECTOR = "#woocommerce_slp_payments_woocommerce_tokenID, #woocommerce_slp_payments_woocommerce_tokenRateUSD, #woocommerce_slp_payments_woocommerce_tokenTickerSymbol," +
    "#woocommerce_slp_payments_woocommerce_tokenName, #woocommerce_slp_payments_woocommerce_tokenImage, #woocommerce_slp_payments_woocommerce_tokenWebsite";

    protected config: SlpPaymentsAdminConfig;
    protected addingNewToken: boolean = false;
    protected selectedToken: string = "";

    constructor(slpPayments: SlpPayments) {
        super(slpPayments);
        this.config = this.slpPayments.window["slpPaymentsAdminCfg"] || {};
        this.config.preInstalledTokens = SlpToken.restoreFromPlainObject(this.config.preInstalledTokens);
        this.config.availableTokens = SlpToken.restoreFromPlainObject(this.config.availableTokens);
    }

    public init() {
        if (this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenID").length === 0)
            return; // not our plugin settings page

        this.slpPayments.getTooltips().initToolTips();
        this.checkReloadPage();
        this.slpPayments.$(".slp-debug-info").attr("disabled", "disabled");
        //this.hideTokensInputRow("#woocommerce_slp_payments_woocommerce_availableTokens");
        this.hideTokensInputRow("#woocommerce_slp_payments_woocommerce_deleteCurrentToken");
        this.hideTokensInputRow("#woocommerce_slp_payments_woocommerce_addingNewToken");
        this.hideTokensInputRow("#woocommerce_slp_payments_woocommerce_allTokenJson");
        this.enableCanEditToken();
        this.addTokenButtonEvents();
        this.addSubmitHandler();
    }

    // ################################################################
    // ###################### PRIVATE FUNCTIONS #######################

    /**
     * Hides the whole table row of an (input) element we only use to send data on form submit.
     * @param inputSelector
     */
    protected hideTokensInputRow(inputSelector: string) {
        if (this.slpPayments.$(inputSelector).length === 0)
            return;
        let parent = this.slpPayments.$(inputSelector).parent();
        const max = 1000;
        let i = 0;
        while (parent.prop("tagName") !== "TR")
        {
            parent = parent.parent();
            i++;
            if (i >= max) {
                this.slpPayments.window.console.error("SlpPayments: Unable to find <tr> tag in WooCommerce settings. Most likely WC HTML changed on update.");
                break;
            }
        }
        parent.addClass("hidden");
    }

    protected addTokenButtonEvents() {
        this.selectedToken = this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenSelect").val();
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenSelect").on("change", (event) => {
            this.updateTokenConfig();
            this.enableCanEditToken();
            this.selectedToken = this.slpPayments.$(event.target).val();
            this.updateTokenInputValues();
        });
        this.slpPayments.$("#slp-add-token").on("click", (event) => {
            if (this.slpPayments.$(event.target).hasClass("disabled") === true)
                return;
            this.addNewToken();
        });
        this.slpPayments.$("#slp-delete-token").on("click", (event) => {
            if (this.slpPayments.$(event.target).hasClass("disabled") === true)
                return;
            if (this.slpPayments.window.confirm(this.config.tr.confirmDeleteToken) === true)
                this.removeToken();
        });
    }

    protected addSubmitHandler() {
        this.slpPayments.$("#mainform").on("submit", (event) => {
            // validate if fields are missing. many fields we can fetch from API via tokenID
            const requiredFields = ["#woocommerce_slp_payments_woocommerce_tokenID", "#woocommerce_slp_payments_woocommerce_tokenRateUSD"];
            for (let i = 0; i < requiredFields.length; i++)
            {
                let value = (this.slpPayments.$(requiredFields[i]).val() || "").trim();
                if (value.length === 0) { // TODO add more validation
                    this.slpPayments.window.alert(this.config.tr.missingFields);
                    event.preventDefault();
                    return;
                }
            }
            if (this.addingNewToken === true)
                localStorage.setItem("reloadPage", "1");

            // must be enabled so that it gets submitted (disabled for pre-installed tokens)
            //this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenTickerSymbol, #woocommerce_slp_payments_woocommerce_addingNewToken").removeAttr("disabled");
            this.slpPayments.$("#mainform input").removeAttr("disabled");
            this.slpPayments.$("#woocommerce_slp_payments_woocommerce_allTokenJson").val(JSON.stringify(this.config.availableTokens));
        });
    }

    protected updateTokenConfig() {
        // if we change the selected token in the dropdown we update all form fields for the newly selected token
        // so we must keep the current config if we switch back
        // on form submission this gets added to the form as JSON to update multiple tokens at once
        let token = {
            id: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenID").val(),
            symbol: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenTickerSymbol").val(),
            name: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenName").val(),
            documentUri: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenWebsite").val(),
            icon: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenImage").val(),
            rate: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenRateUSD").val(),
            decimals: 0, // crawled from blockchain
            enabled: this.slpPayments.$("#woocommerce_slp_payments_woocommerce_enableCurrentToken").is(":checked") === true
        };
        for (let prop in token)
        {
            if (typeof token[prop] === "string" && token[prop].trim().length === 0)
                return; // discard changes if values are empty to prevent accidental overwrites
        }
        token.rate = parseFloat(token.rate);
        if (token.rate === Number.NaN)
            return;
        this.config.availableTokens[this.selectedToken] = Object.assign(new SlpToken(), token); // add or overwrite
    }

    protected enableCanEditToken() {
        const preInstalledTokenSymbols = Object.keys(this.config.preInstalledTokens);
        const selectedToken = this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenSelect").val();
        const tokenControlSelector = "#slp-delete-token";
        if (preInstalledTokenSymbols.indexOf(selectedToken) !== -1) {
            this.slpPayments.$(tokenControlSelector).addClass("disabled");
            this.slpPayments.$(AdminControls.TOKEN_INPUT_SELECTOR).attr("disabled", "disabled");
        }
        else {
            this.slpPayments.$(tokenControlSelector).removeClass("disabled");
            this.slpPayments.$(AdminControls.TOKEN_INPUT_SELECTOR).removeAttr("disabled");
        }
    }

    protected updateTokenInputValues() {
        const selectedTickerSymbol = this.selectedToken;
        const selectedToken = this.config.availableTokens[selectedTickerSymbol];
        if (!selectedToken) {
            this.slpPayments.$(AdminControls.TOKEN_INPUT_SELECTOR).val(""); // a new token not in our database yet
            return;
        }
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_enableCurrentToken").prop("checked", selectedToken.enabled === true);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenID").val(selectedToken.id);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenRateUSD").val(selectedToken.rate);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenTickerSymbol").val(selectedToken.symbol);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenName").val(selectedToken.name);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenImage").val(selectedToken.icon);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenWebsite").val(selectedToken.documentUri);
    }

    protected addNewToken() {
        this.slpPayments.$(AdminControls.TOKEN_INPUT_SELECTOR).val("");
        const tokenSelect = this.slpPayments.$("#woocommerce_slp_payments_woocommerce_tokenSelect");
        tokenSelect.append(new Option(this.config.tr.newToken, "newToken", true, true));
        tokenSelect.find("option[value='newToken']").prop("selected", true);
        this.slpPayments.$("#woocommerce_slp_payments_woocommerce_addingNewToken").val("1");
        this.enableCanEditToken();
        this.slpPayments.$("#slp-delete-token").addClass("disabled");
        this.addingNewToken = true;
    }

    protected removeToken() {
        const form = this.slpPayments.$("#mainform");
        /* // must be present in PHP output for WC to parse it (easier)
        this.slpPayments.$("<input>", {
            type: "hidden",
            name: "woocommerce_slp_payments_woocommerce_delete_token",
            value: "1"
        }).appendTo(form);*/
        //form.find("#woocommerce_slp_payments_woocommerce_deleteCurrentToken").val("1").removeAttr("disabled");
        localStorage.setItem("reloadPage", "1");
        form.find("button[name='save']").click(); // submit only processed if button is clicked (not with submit() event)
    }

    protected checkReloadPage() {
        // when our new token settings are saved, our WP hooks already generate the current html output
        // So when we create a new token it's not present. easy solution: reload
        // TODO better solution would be to submit new token via WP REST API instead of form submit()
        // TODO add some "reloading" message via JS
        if (localStorage.getItem("reloadPage") !== "1")
            return;
        localStorage.removeItem("reloadPage");
        setTimeout(() => {
            this.slpPayments.window.document.location.reload();
        }, 10);
    }
}
